from dataclasses import dataclass
from typing import (
    Annotated,
    Any,
)

from pydantic import BaseModel
from pydantic.functional_validators import AfterValidator

from exasol.ai.mcp.server.utils import sql_text_value


def check_no_double_quotes(v: str) -> str:
    if '"' in v:
        raise ValueError("Double-quote characters are not allowed in a field name.")
    return v


NoDoubleQuotesStr = Annotated[str, AfterValidator(check_no_double_quotes)]


@dataclass
class ExaDbResult:
    result: list[str:Any]


class MetaSettings(BaseModel):
    """
    The general settings for a single type of metadata, e.g. tables.
    """

    enable: bool = True
    """
    Allows to disable the listing of a particular type of metadata.
    """

    name_field: NoDoubleQuotesStr = "name"
    """
    The name of the output field that contains the object name, e.g. "table_name".
    """

    comment_field: NoDoubleQuotesStr = "comment"
    """
    The name of the output field that contains the comment, e.g. "table_comment".
    """


class MetaListSettings(MetaSettings):
    """
    The settings for a type of metadata that can be listed and filtered.
    """

    like_pattern: str | None = None
    """
    An optional sql-style pattern for the object name filtering.

    Use case example: The user wants to create a set of purified de-normalised views on
    the existing database and limit the table listings to only these views. One way of
    achieving this is to create the views in a new schema and limit the listing of the
    schemas to this schema only. In the case of no permission to create schema, one can
    create the views in an existing schema and use some prefix for name disambiguation.
    This prefix can also be used for filtering the views in the listing.
    """
    regexp_pattern: str | None = None
    """
    An optional regular expression pattern for the object name filtering.
    Both like_pattern and regexp_pattern can be used at the same time, although there is
    not much point in doing so.
    """

    @property
    def select_predicate(self) -> str:
        """
        The SQL predicate for the object filtering by name.
        Empty string if neigher of the filtering patterns are defined.
        """
        conditions: list[str] = []
        if self.like_pattern:
            conditions.append(
                f"""local."{self.name_field}" LIKE {sql_text_value(self.like_pattern)}"""
            )
        if self.regexp_pattern:
            conditions.append(
                f"""local."{self.name_field}" REGEXP_LIKE {sql_text_value(self.regexp_pattern)}"""
            )
        return " AND ".join(conditions)


class MetaColumnSettings(MetaSettings):
    """
    The settings for listing columns and constraints when describing a table.
    """

    type_field: NoDoubleQuotesStr = "type"
    """
    The name of the output field for the column SQL type.
    """

    constraint_name_field: NoDoubleQuotesStr = "name"
    """
    The name of the output field for the constraint name if it was specified.
    """

    constraint_type_field: NoDoubleQuotesStr = "type"
    """
    The name of the output field for the constraint type, e.g. PRIMARY KEY.
    """

    constraint_columns_field: NoDoubleQuotesStr = "columns"
    """
    The name of the output field for the list of columns the constraint is applied to.
    """

    referenced_schema_field: NoDoubleQuotesStr = "referenced_schema"
    """
    The name of the output field for the referenced schema in the FOREIGN KEY constraint.
    """

    referenced_table_field: NoDoubleQuotesStr = "referenced_table"
    """
    The name of the output field for the referenced table in the FOREIGN KEY constraint.
    """

    referenced_columns_field: NoDoubleQuotesStr = "referenced_columns"
    """
    The name of the output field for the list of columns in a referenced table in the
    FOREIGN KEY constraint.
    """

    columns_field: NoDoubleQuotesStr = "columns"
    """
    The name of the output field for the list of columns in a table being described.
    """

    constraints_field: NoDoubleQuotesStr = "constraints"
    """
    The name of the output field for the list of constraints in a table being described.
    """

    table_comment_field: NoDoubleQuotesStr = "table_comment"
    """
    The name of the output field for the table/view comment.
    """

    usage_field: NoDoubleQuotesStr = "usage"
    """
    The name of the output field for general instructions on using a table.
    """


class MetaParameterSettings(MetaSettings):
    """
    The settings for listing input/output parameters when describing a function or a
    UDF script.
    """

    type_field: NoDoubleQuotesStr = "parameter_type"
    """
    The name of the output field for the SQL type of a parameter or return value.
    """

    input_field: NoDoubleQuotesStr = "inputs"
    """
    The name of the output field for the list of input parameters.
    """

    return_field: NoDoubleQuotesStr = "returns"
    """
    The name of the output field for the return value.
    """

    emit_field: NoDoubleQuotesStr = "emits"
    """
    The name of the output field for the list of parameters emitted by a UDF.
    """

    usage_field: NoDoubleQuotesStr = "usage"
    """
    The name of the output field for a function or UDF usage, e.g. a call example.
    """


class McpServerSettings(BaseModel):
    """
    MCP server configuration.
    """

    schemas: MetaListSettings = MetaListSettings(
        name_field="schema_name", comment_field="schema_comment"
    )
    tables: MetaListSettings = MetaListSettings(
        name_field="table_name", comment_field="table_comment"
    )
    views: MetaListSettings = MetaListSettings(
        enable=False, name_field="table_name", comment_field="table_comment"
    )
    functions: MetaListSettings = MetaListSettings(
        name_field="function_name", comment_field="function_comment"
    )
    scripts: MetaListSettings = MetaListSettings(
        name_field="script_name", comment_field="script_comment"
    )
    columns: MetaColumnSettings = MetaColumnSettings(
        name_field="column_name", comment_field="column_comment"
    )
    parameters: MetaParameterSettings = MetaParameterSettings(
        name_field="parameter_name", comment_field="function_comment"
    )
    enable_read_query: bool = False



import json
import os
import re
from functools import cache
from textwrap import dedent
from typing import (
    Annotated,
    Any,
)

import pyexasol
from fastmcp import FastMCP
from pydantic import (
    Field,
    ValidationError,
)
from sqlglot import (
    exp,
    parse_one,
)
from sqlglot.errors import ParseError

from exasol.ai.mcp.server.parameter_parser import (
    FuncParameterParser,
    ScriptParameterParser,
)
from exasol.ai.mcp.server.parameter_pattern import (
    parameter_list_pattern,
    regex_flags,
)
from exasol.ai.mcp.server.server_settings import (
    ExaDbResult,
    McpServerSettings,
    MetaListSettings,
)
from exasol.ai.mcp.server.utils import sql_text_value

ENV_SETTINGS = "EXA_MCP_SETTINGS"
ENV_DSN = "EXA_DSN"
ENV_USER = "EXA_USER"
ENV_PASSWORD = "EXA_PASSWORD"

TABLE_USAGE = (
    "In an SQL query, the names of database objects, such as schemas, "
    "tables and columns should be enclosed in double quotes. "
    "A reference to a table should include a reference to its schema. "
    "The SELECT column list cannot have both the * and explicit column names. "
)


def _where_clause(*predicates) -> str:
    condition = " AND ".join(filter(bool, predicates))
    if condition:
        return f"WHERE {condition}"
    return ""


def _validate_describe_table_args(schema_name: str, table_name: str) -> None:
    if not schema_name:
        raise ValueError("Schema name is not provided.")
    if not table_name:
        raise ValueError("Table name is not provided.")


@cache
def _get_emits_pattern() -> re.Pattern:
    pattern = rf"EMITS\s*\({parameter_list_pattern}\)"
    return re.compile(pattern, flags=regex_flags)


def verify_query(query: str) -> bool:
    """
    Verifies that the query is a valid SELECT query.
    Declines any other types of statements including the SELECT INTO.
    """

    # Here is a fix for the SQLGlot deficiency in understanding the syntax of variadic
    # emit UDF. The EMITS clause in the SELECT statement is currently not recognised.
    # To let SQLGlot validate the query, this clause must be pinched away.
    query = _get_emits_pattern().sub("", query)

    try:
        ast = parse_one(query, read="exasol")
        if isinstance(ast, exp.Select):
            return "into" not in ast.args
        return False
    except ParseError:
        return False


class ExasolMCPServer(FastMCP):
    """
    An Exasol MCP server based on FastMCP.

    Args:
        connection:
            pyexasol connection object. Note: the connection should be created
            with `fetch_dict`=True. The server sets this option to True anyway.
        config:
            The server configuration.
    """

    def __init__(
        self, connection: pyexasol.ExaConnection, config: McpServerSettings
    ) -> None:
        super().__init__(name="exasol-mcp")
        self.connection = connection
        self.connection.options["fetch_dict"] = True
        self.config = config
        self._register_tools()

    def _register_tools(self):
        if self.config.schemas.enable:
            self.tool(
                self.list_schemas,
                description=(
                    "The tool lists schemas in the Exasol Database. "
                    "For each schema provides the name and an optional comment."
                ),
            )
        if self.config.tables.enable or self.config.views.enable:
            self.tool(
                self.list_tables,
                description=(
                    "The tool lists tables in the specified schema of the Exasol "
                    "Database. For each table it provides the name and an optional "
                    "comment."
                ),
            )
        if self.config.functions.enable:
            self.tool(
                self.list_functions,
                description=(
                    "The tool lists functions in the specified schema of the Exasol "
                    "Database. For each function it provides the name and an optional "
                    "comment."
                ),
            )
        if self.config.scripts.enable:
            self.tool(
                self.list_scripts,
                description=(
                    "The tool lists the user defined functions (UDF) in the specified "
                    "schema of the Exasol Database. For each function it provides the "
                    "name and an optional comment."
                ),
            )
        if self.config.columns.enable:
            self.tool(
                self.describe_table,
                description=(
                    "The tool describes the specified table in the specified schema of "
                    "the Exasol Database. The description includes the list of columns "
                    "and the list of constraints. For each column the tool provides "
                    "the name, the SQL data type and an optional comment. For each "
                    "constraint it provides its type, e.g. PRIMARY KEY, the list of "
                    "columns the constraint is applied to and an optional name. For a "
                    "FOREIGN KEY it also provides the referenced schema, table and a "
                    "list of columns in the referenced table."
                ),
            )
        if self.config.parameters.enable:
            self.tool(
                self.describe_function,
                description=(
                    "The tool describes the specified function in the specified schema "
                    "of the Exasol Database. It provides the list of input parameters "
                    "and the return SQL type. For each parameter it specifies the name "
                    "and the SQL type."
                ),
            )
            self.tool(
                self.describe_script,
                description=(
                    "The tool describes the specified user defined function in the "
                    "specified schema of the Exasol Database. It provides the list of "
                    "input parameters, the list of emitted parameters or the SQL type "
                    "of a single returned value. For each parameter it provides the "
                    "name and the SQL type. Both the input and the emitted parameters "
                    "can be dynamic, or, in other words, flexible. The dynamic parameters "
                    "are indicated with ... (triple dot) string instead of the parameter "
                    "list. A user defined function with dynamic input parameters can "
                    "be called using the same syntax as a normal function. If the "
                    "function output is emitted dynamically, the list of output "
                    "parameters must be supplied in the call. This can be achieved by "
                    'appending the select statement with the special term "emits" '
                    "followed by the parameter list in parentheses."
                ),
            )
        if self.config.enable_read_query:
            self.tool(
                self.execute_query,
                description=(
                    "The tool executes the specified query in the specified schema of "
                    "the Exasol Database. The query must be a SELECT statement. The "
                    "tool returns data selected by the query."
                ),
            )

    def _build_meta_query(
        self, meta_name: str, conf: MetaListSettings, schema_name: str, *predicates
    ) -> str:
        """
        Builds a metadata query.

        Args:
            meta_name:
                Must be one of "SCHEMA", "TABLE", "VIEW", "FUNCTION", or "SCRIPT".
            conf:
                Metadata type settings, which is a part of the server configuration.
            schema_name:
                The schema name provided in the call to the tool. Ignored if meta_name
                =='SCHEMA'. Otherwise, if it's empty the current schema of the pyexasol
                connection may be used. If there is no current schema, the query will
                return objects from all schemas.
            predicates:
                Any additional predicates to be used in the WHERE clause.
        """
        predicates = [conf.select_predicate, *predicates]
        if meta_name != "SCHEMA":
            schema_name = schema_name or self.connection.current_schema()
            if schema_name:
                predicates.append(f"{meta_name}_SCHEMA = {sql_text_value(schema_name)}")
        return dedent(
            f"""
            SELECT {meta_name}_NAME AS "{conf.name_field}", {meta_name}_COMMENT AS "{conf.comment_field}"
            FROM SYS.EXA_ALL_{meta_name}S
            {_where_clause(*predicates)}
        """
        )

    def _execute_meta_query(self, query: str) -> ExaDbResult:
        result = self.connection.meta.execute_snapshot(query=query).fetchall()
        return ExaDbResult(result)

    def list_schemas(self) -> ExaDbResult:
        conf = self.config.schemas
        if not conf.enable:
            raise RuntimeError("Schema listing is disabled.")

        query = self._build_meta_query("SCHEMA", conf, "")
        return self._execute_meta_query(query)

    def list_tables(
        self,
        schema_name: Annotated[
            str, Field(description="name of the database schema", default="")
        ],
    ) -> ExaDbResult:
        table_conf = self.config.tables
        view_conf = self.config.views
        if (not table_conf.enable) and (not view_conf.enable):
            raise RuntimeError("Both table and view listings are disabled.")

        query = "\nUNION\n".join(
            self._build_meta_query(meta_name, conf, schema_name)
            for meta_name, conf in zip(["TABLE", "VIEW"], [table_conf, view_conf])
            if conf.enable
        )
        return self._execute_meta_query(query)

    def list_functions(
        self,
        schema_name: Annotated[
            str, Field(description="name of the database schema", default="")
        ],
    ) -> ExaDbResult:
        conf = self.config.functions
        if not conf.enable:
            raise RuntimeError("Function listing is disabled.")

        query = self._build_meta_query("FUNCTION", conf, schema_name)
        return self._execute_meta_query(query)

    def list_scripts(
        self,
        schema_name: Annotated[
            str, Field(description="name of the database schema", default="")
        ],
    ) -> ExaDbResult:
        conf = self.config.scripts
        if not conf.enable:
            raise RuntimeError("Script listing is disabled.")

        query = self._build_meta_query(
            "SCRIPT", conf, schema_name, "SCRIPT_TYPE = 'UDF'"
        )
        return self._execute_meta_query(query)

    def describe_columns(
        self,
        schema_name: Annotated[
            str, Field(description="name of the database schema", default="")
        ],
        table_name: Annotated[str, Field(description="name of the table", default="")],
    ) -> ExaDbResult:
        """
        Returns the list of columns in the given table. Currently, this is a part of
        the `describe_table` tool, but it can be used independently in the future.
        """
        conf = self.config.columns
        if not conf.enable:
            raise RuntimeError("Column listing is disabled.")
        schema_name = schema_name or self.connection.current_schema()
        _validate_describe_table_args(schema_name, table_name)

        query = dedent(
            f"""
            SELECT
                COLUMN_NAME AS "{conf.name_field}",
                COLUMN_TYPE AS "{conf.type_field}",
                COLUMN_COMMENT AS "{conf.comment_field}"
            FROM SYS.EXA_ALL_COLUMNS
            WHERE
                COLUMN_SCHEMA = {sql_text_value(schema_name)} AND
                COLUMN_TABLE = {sql_text_value(table_name)}
        """
        )
        return self._execute_meta_query(query)

    def describe_constraints(
        self,
        schema_name: Annotated[
            str, Field(description="name of the database schema", default="")
        ],
        table_name: Annotated[str, Field(description="name of the table", default="")],
    ) -> ExaDbResult:
        """
        Returns the list of constraints in the given table. Currently, this is a part
        of the `describe_table` tool, but it can be used independently in the future.
        """
        conf = self.config.columns
        if not conf.enable:
            raise RuntimeError("Constraint listing is disabled.")
        schema_name = schema_name or self.connection.current_schema()
        _validate_describe_table_args(schema_name, table_name)

        query = dedent(
            f"""
            SELECT
                FIRST_VALUE(CONSTRAINT_TYPE) AS "{conf.constraint_type_field}",
                CASE LEFT(CONSTRAINT_NAME, 4) WHEN 'SYS_' THEN NULL
                    ELSE CONSTRAINT_NAME END AS "{conf.constraint_name_field}",
                GROUP_CONCAT(DISTINCT COLUMN_NAME ORDER BY ORDINAL_POSITION)
                    AS "{conf.constraint_columns_field}",
                FIRST_VALUE(REFERENCED_SCHEMA) AS "{conf.referenced_schema_field}",
                FIRST_VALUE(REFERENCED_TABLE) AS "{conf.referenced_table_field}",
                GROUP_CONCAT(DISTINCT REFERENCED_COLUMN ORDER BY ORDINAL_POSITION)
                    AS "{conf.referenced_columns_field}"
            FROM SYS.EXA_ALL_CONSTRAINT_COLUMNS
            WHERE
                CONSTRAINT_SCHEMA = {sql_text_value(schema_name)} AND
                CONSTRAINT_TABLE = {sql_text_value(table_name)}
            GROUP BY CONSTRAINT_NAME
        """
        )
        return self._execute_meta_query(query)

    def get_table_comment(self, schema_name: str, table_name: str) -> str | None:
        schema_name = schema_name or self.connection.current_schema()
        # `table_name` can be the name of a table or a view.
        # This query tries both possibilities. The UNION clause collapses
        # the result into a single non-NULL distinct value.
        query = dedent(
            f"""
            SELECT TABLE_COMMENT AS COMMENT FROM SYS.EXA_ALL_TABLES
            WHERE
                TABLE_SCHEMA = {sql_text_value(schema_name)} AND
                TABLE_NAME = {sql_text_value(table_name)}
            UNION
            SELECT VIEW_COMMENT AS COMMENT FROM SYS.EXA_ALL_VIEWS
            WHERE
                VIEW_SCHEMA = {sql_text_value(schema_name)} AND
                VIEW_NAME = {sql_text_value(table_name)}
            LIMIT 1;
        """
        )
        comment_row = self.connection.meta.execute_snapshot(query=query).fetchone()
        if comment_row is None:
            return None
        table_comment = next(iter(comment_row.values()))
        if table_comment is None:
            return None
        return str(table_comment)

    def describe_table(
        self,
        schema_name: Annotated[
            str, Field(description="name of the database schema", default="")
        ],
        table_name: Annotated[str, Field(description="name of the table", default="")],
    ) -> dict[str, Any]:

        conf = self.config.columns
        columns = self.describe_columns(schema_name, table_name)
        constraints = self.describe_constraints(schema_name, table_name)
        table_comment = self.get_table_comment(schema_name, table_name)

        return {
            conf.columns_field: columns.result,
            conf.constraints_field: constraints.result,
            conf.table_comment_field: table_comment,
            conf.usage_field: TABLE_USAGE,
        }

    def describe_function(
        self,
        schema_name: Annotated[
            str, Field(description="name of the database schema", default="")
        ],
        func_name: Annotated[
            str, Field(description="name of the function", default="")
        ],
    ) -> dict[str, Any]:
        parser = FuncParameterParser(
            connection=self.connection, conf=self.config.parameters
        )
        return parser.describe(schema_name, func_name)

    def describe_script(
        self,
        schema_name: Annotated[
            str, Field(description="name of the database schema", default="")
        ],
        script_name: Annotated[
            str, Field(description="name of the script", default="")
        ],
    ) -> dict[str, Any]:
        parser = ScriptParameterParser(
            connection=self.connection, conf=self.config.parameters
        )
        return parser.describe(schema_name, script_name)

    def execute_query(
        self, query: Annotated[str, Field(description="select query")]
    ) -> ExaDbResult:
        if not self.config.enable_read_query:
            raise RuntimeError("Query execution is disabled.")
        if verify_query(query):
            result = self.connection.execute(query=query).fetchall()
            return ExaDbResult(result)
        raise ValueError("The query is invalid or not a SELECT statement.")


def get_mcp_settings() -> McpServerSettings:
    """
    Reads optional settings. They can be provided either in a json string stored in the
    EXA_MCP_SETTINGS environment variable or in a json file. In the latter case
    EXA_MCP_SETTINGS must contain the file path.
    """
    try:
        settings_text = os.environ.get(ENV_SETTINGS)
        if not settings_text:
            return McpServerSettings()
        elif re.match(r"^\s*\{.*\}\s*$", settings_text):
            return McpServerSettings.model_validate_json(settings_text)
        elif os.path.isfile(settings_text):
            with open(settings_text) as f:
                return McpServerSettings.model_validate(json.load(f))
        raise ValueError(
            "Invalid MCP Server configuration settings. The configuration "
            "environment variable should either contain a json string or "
            "point to an existing json file."
        )
    except (ValidationError, json.decoder.JSONDecodeError) as config_error:
        raise ValueError("Invalid MCP Server configuration settings.") from config_error


def main():
    """
    Main entry point that creates and runs the MCP server.
    """
    dsn = os.environ[ENV_DSN]
    user = os.environ[ENV_USER]
    password = os.environ[ENV_PASSWORD]
    mcp_settings = get_mcp_settings()

    conn = pyexasol.connect(
        dsn=dsn, user=user, password=password, fetch_dict=True, compression=True
    )

    mcp_server = ExasolMCPServer(connection=conn, config=mcp_settings)
    mcp_server.run()


if __name__ == "__main__":
    main()



import re
from abc import (
    ABC,
    abstractmethod,
)
from textwrap import dedent
from typing import Any

from pyexasol import ExaConnection

from exasol.ai.mcp.server.parameter_pattern import (
    exa_type_pattern,
    identifier_pattern,
    parameter_list_pattern,
    quoted_identifier_pattern,
    regex_flags,
)
from exasol.ai.mcp.server.server_settings import MetaParameterSettings
from exasol.ai.mcp.server.utils import sql_text_value

VARIADIC_MARKER = "..."

FUNCTION_USAGE = (
    "In an SQL query, the names of database objects, such as schemas, tables, "
    "functions, and columns should be enclosed in double quotes. "
    "A reference to a function should include a reference to its schema."
)


class ParameterParser(ABC):
    def __init__(self, connection: ExaConnection, conf: MetaParameterSettings) -> None:
        self.connection = connection
        self.conf = conf
        self._parameter_extract_pattern: re.Pattern | None = None

    def _execute_query(self, query: str) -> list[dict[str, Any]]:
        return self.connection.meta.execute_snapshot(query=query).fetchall()

    def describe(
        self,
        schema_name: str,
        func_name: str,
    ) -> dict[str, Any]:
        """
        Requests and parses metadata for the specified function or script.
        """
        if not self.conf.enable:
            raise RuntimeError("Parameter listing is disabled.")
        schema_name = schema_name or self.connection.current_schema()
        if not schema_name:
            raise ValueError("Schema name is not provided.")
        if not func_name:
            raise ValueError("Function or script name is not provided.")

        query = self.get_func_query(schema_name, func_name)
        result = self._execute_query(query=query)
        if not result:
            raise ValueError(
                f"The function or script {schema_name}.{func_name} not found."
            )
        script_info = result[0]
        return self.extract_parameters(script_info)

    @property
    def parameter_extract_pattern(self) -> re.Pattern:
        r"""
        Lazily compiles a pattern for extracting parameter names and their SQL types
        from a parameter list.

        The pattern is looking for a sequence of parameters, each starting either
        from the beginning or from comma: (?:^|,). The parameter should be followed
        by either the end of the input or comma: (?=\Z|,).
        """
        if self._parameter_extract_pattern is not None:
            return self._parameter_extract_pattern

        pattern = (
            rf"(?:^|,)\s*(?P<{self.conf.name_field}>{quoted_identifier_pattern})"
            rf"\s+(?P<{self.conf.type_field}>{exa_type_pattern})\s*(?=\Z|,)"
        )
        self._parameter_extract_pattern = re.compile(pattern, flags=regex_flags)
        return self._parameter_extract_pattern

    def parse_parameter_list(self, params: str) -> str | list[dict[str, str]]:
        """
        Breaks the input string into parameter definitions. The input string should be
        extracted from a text of a function or a script and contain a list of parameters,
        where each parameter consists of a name and an SQL type. The list can be either
        an input or, in case of an EMIT UDF, the emit list.

        The parameters may be dynamic (variadic). This is designated with "...". In
        such case, the function simply returns "...".

        Otherwise, the function returns a list of dictionaries, where each dictionary
        describes a parameter and includes its name and type. The dictionary keys are
        defined by the provided configuration. The double quotes in the parameter name
        gets removed.
        """

        def remove_double_quotes(di: dict[str, str]) -> dict[str, str]:
            return {key: val.strip('"') for key, val in di.items()}

        params = params.lstrip()
        if params == VARIADIC_MARKER:
            return params

        return [
            remove_double_quotes(m.groupdict())
            for m in self.parameter_extract_pattern.finditer(params)
        ]

    def format_return_type(self, param: str) -> str | dict[str, str]:
        return {self.conf.type_field: param}

    @abstractmethod
    def get_func_query(self, schema_name: str, func_name: str) -> str:
        """
        Builds a query requesting metadata for a given function or script.
        """

    @abstractmethod
    def extract_parameters(self, info: dict[str, Any]) -> dict[str, Any]:
        """
        Parses the text of a function or a UDF script, extracting its input and output
        parameters and/or return type. Returns the result in a json form.

        Should raise a ValueError if the parsing fails.

        Note: This function does not validate the entire function or script text. It is
        only looking at its header.

        Args:
            info:
                The function or script information obtained from reading the relevant
                row in respectively SYS.EXA_ALL_FUNCTIONS or SYS.EXA_ALL_SCRIPTS table.
        """


class FuncParameterParser(ParameterParser):

    def __init__(self, connection: ExaConnection, conf: MetaParameterSettings) -> None:
        super().__init__(connection, conf)
        self._func_pattern: re.Pattern | None = None

    def get_func_query(self, schema_name: str, func_name: str) -> str:
        return dedent(
            f"""
            SELECT * FROM SYS.EXA_ALL_FUNCTIONS
            WHERE FUNCTION_SCHEMA = {sql_text_value(schema_name)} AND
                FUNCTION_NAME = {sql_text_value(func_name)}
        """
        )

    @property
    def func_pattern(self) -> re.Pattern:
        """
        Lazily compiles the function parsing pattern.
        """
        if self._func_pattern is not None:
            return self._func_pattern

        # The schema is optional
        func_schema_pattern = rf"(?:{quoted_identifier_pattern}\s*\.\s*)?"
        func_name_pattern = quoted_identifier_pattern
        pattern = (
            r"\A\s*FUNCTION\s+"
            rf"{func_schema_pattern}{func_name_pattern}\s*"
            rf"\((?P<{self.conf.input_field}>{parameter_list_pattern})\)\s*"
            rf"RETURN\s+(?P<{self.conf.return_field}>{exa_type_pattern})\s+"
        )
        self._func_pattern = re.compile(pattern, flags=regex_flags)
        return self._func_pattern

    def extract_parameters(self, info: dict[str, Any]) -> dict[str, Any]:
        m = self.func_pattern.match(info["FUNCTION_TEXT"])
        if m is None:
            raise ValueError(
                "Failed to parse the text of the function "
                f'{info["FUNCTION_SCHEMA"]}.{info["FUNCTION_NAME"]}.'
            )
        return {
            self.conf.input_field: self.parse_parameter_list(
                m.group(self.conf.input_field)
            ),
            self.conf.return_field: self.format_return_type(
                m.group(self.conf.return_field)
            ),
            self.conf.comment_field: info["FUNCTION_COMMENT"],
            self.conf.usage_field: FUNCTION_USAGE,
        }


class ScriptParameterParser(ParameterParser):

    def __init__(self, connection: ExaConnection, conf: MetaParameterSettings) -> None:
        super().__init__(connection, conf)
        self._emit_pattern: re.Pattern | None = None
        self._return_pattern: re.Pattern | None = None

    def get_func_query(self, schema_name: str, func_name: str) -> str:
        return dedent(
            f"""
            SELECT * FROM SYS.EXA_ALL_SCRIPTS
            WHERE SCRIPT_SCHEMA = {sql_text_value(schema_name)} AND
                SCRIPT_NAME = {sql_text_value(func_name)}
        """
        )

    def _udf_pattern(self, emits: bool) -> re.Pattern:
        """Compiles a pattern for parsing a UDF script."""

        # The parameter matching pattern should account for the possibility of the
        # variadic syntax: ...
        dynamic_list_pattern = rf"(?:\s*...\s*|{parameter_list_pattern})"

        output_pattern = (
            rf"EMITS\s*\((?P<{self.conf.emit_field}>{dynamic_list_pattern})\)\s*"
            if emits
            else rf"RETURNS\s+(?P<{self.conf.return_field}>{exa_type_pattern})\s+"
        )
        language_pattern = identifier_pattern
        # The schema is optional.
        udf_schema_pattern = rf"(?:{quoted_identifier_pattern}\s*\.\s*)?"
        udf_name_pattern = quoted_identifier_pattern

        pattern = (
            rf"\A\s*CREATE\s+{language_pattern}\s+(?:SCALAR|SET)\s+SCRIPT\s+"
            rf"{udf_schema_pattern}{udf_name_pattern}\s*"
            rf"\((?P<{self.conf.input_field}>{dynamic_list_pattern})\)\s*"
            rf"{output_pattern}AS\s+"
        )
        return re.compile(pattern, flags=regex_flags)

    @property
    def emit_udf_pattern(self) -> re.Pattern:
        """
        Lazily compiles the Emit-UDF parsing pattern.
        """
        if self._emit_pattern is None:
            self._emit_pattern = self._udf_pattern(emits=True)
        return self._emit_pattern

    @property
    def return_udf_pattern(self) -> re.Pattern:
        """
        Lazily compiles the Return-UDF parsing pattern.
        """
        if self._return_pattern is None:
            self._return_pattern = self._udf_pattern(emits=False)
        return self._return_pattern

    def extract_udf_parameters(self, info: dict[str, Any]) -> dict[str, Any]:
        pattern = (
            self.emit_udf_pattern
            if info["SCRIPT_RESULT_TYPE"] == "EMITS"
            else self.return_udf_pattern
        )
        m = pattern.match(info["SCRIPT_TEXT"])
        if m is None:
            raise ValueError(
                "Failed to parse the text of the UDF script "
                f'{info["SCRIPT_SCHEMA"]}.{info["SCRIPT_NAME"]}.'
            )
        param_func = {
            self.conf.input_field: self.parse_parameter_list,
            self.conf.emit_field: self.parse_parameter_list,
            self.conf.return_field: self.format_return_type,
        }
        return {
            param_field: param_func[param_field](params)
            for param_field, params in m.groupdict().items()
            if params or (param_field == self.conf.input_field)
        }

    def _get_variadic_note(self, variadic_input: bool, variadic_emit: bool) -> str:
        """
        A helper function for generating code example. Writes an explanation of the
        variadic syntax.
        """
        if (not variadic_input) and (not variadic_emit):
            return ""
        if not variadic_emit:
            variadic_param = "input"
            variadic_action = "provided"
        elif not variadic_input:
            variadic_param = "output"
            variadic_action = "emitted"
        else:
            variadic_param = "input and output"
            variadic_action = "provided and emitted"
        variadic_emit_note = (
            (
                " When calling a UDF with dynamic output parameters, the EMITS clause "
                "should be provided in the call, as demonstrated in the example below."
            )
            if variadic_emit
            else ""
        )
        return (
            f" This particular UDF has dynamic {variadic_param} parameters. "
            f"The {self.conf.comment_field} may give a hint on what parameters are "
            f"expected to be {variadic_action} in a specific use case."
            f"{variadic_emit_note} Note that in the following example the "
            f"{variadic_param} parameters are given only for illustration. They "
            f"shall not be used as a guide on how to call this UDF."
        )

    @staticmethod
    def _get_emit_note(emit: bool, emit_size: int, func_type: str) -> str:
        """
        A helper function for generating code example. Writes an explanation of the
        difference between a normal function and an EMIT UDF.
        """
        if not emit:
            return ""
        input_unit = "row" if func_type == "scalar" else "group"
        if emit_size == 0:
            output_desc = ""
        elif emit_size == 1:
            output_desc = ", one column each"
        else:
            output_desc = f", each with {emit_size} columns"
        return (
            f" Unlike normal {func_type} functions that return a single value for "
            f"every input {input_unit}, this UDF can emit multiple output rows per "
            f"input {input_unit}{output_desc}. An SQL SELECT statement calling a UDF "
            f"that emits output columns, such as this one, cannot include any "
            f"additional columns."
        )

    @staticmethod
    def _get_general_note(emit: bool) -> str:
        emit_note = ", including columns returned by the UDF," if emit else ""
        return (
            "Note that in an SQL query, the names of database objects, such as "
            f"schemas, tables, UDFs, and columns{emit_note} should be "
            "enclosed in double quotes. "
            "A reference to a UDF should include a reference to its schema."
        )

    def get_udf_call_example(
        self, result: dict[str, Any], input_type: str, func_name: str
    ) -> str:
        """
        Generates call example for a given UDF. For the examples of the
        generated texts see `test_get_udf_call_example` unit test.
        """
        emit = self.conf.emit_field in result
        variadic_emit = emit and result[self.conf.emit_field] == VARIADIC_MARKER
        variadic_input = result[self.conf.input_field] == VARIADIC_MARKER
        emit_size = (
            len(result[self.conf.emit_field]) if emit and (not variadic_emit) else 0
        )
        func_type = "scalar" if input_type.upper() == "SCALAR" else "aggregate"
        if variadic_input:
            input_params = '"INPUT_1", "INPUT_2"'
        else:
            input_params = ", ".join(
                f'"{param[self.conf.name_field]}"'
                for param in result[self.conf.input_field]
            )
        if variadic_emit:
            output_params = '"OUTPUT_1", "OUTPUT_2"'
        elif emit:
            output_params = ", ".join(
                f'"{param[self.conf.name_field]}"'
                for param in result[self.conf.emit_field]
            )
        else:
            output_params = ""

        introduction = (
            f"In most cases, an Exasol {input_type} User Defined Function (UDF) can "
            f"be called just like a normal {func_type} function."
            f"{self._get_emit_note(emit, emit_size, func_type)}"
            f"{self._get_variadic_note(variadic_input, variadic_emit)}"
        )

        example_header = "Here is a usage example for this particular UDF:"

        return_alias = "" if emit else ' AS "RETURN_VALUE"'
        emit_clause = (
            ' EMITS ("OUTPUT_1" VARCHAR(100), "OUTPUT_2" DOUBLE)'
            if variadic_emit
            else ""
        )
        from_clause = ' FROM "MY_SOURCE_TABLE"' if input_params else ""
        example = (
            f'```\nSELECT "{func_name}"({input_params}){return_alias}{emit_clause}'
            f"{from_clause}\n```"
        )

        example_footer = (
            (
                "This example assumes that the currently opened schema has the table "
                f'"MY_SOURCE_TABLE" with the following columns: {input_params}.'
            )
            if input_params
            else ""
        )
        if emit:
            emit_note = (
                f"The query produces a result set with the columns ({output_params}), "
                "similar to what is returned by a SELECT query."
            )
            example_footer = f"{example_footer}\n{emit_note}"

        return "\n".join(
            [
                introduction,
                example_header,
                example,
                example_footer,
                self._get_general_note(emit),
            ]
        )

    def extract_parameters(self, info: dict[str, Any]) -> dict[str, Any]:
        result = self.extract_udf_parameters(info)
        result[self.conf.comment_field] = info["SCRIPT_COMMENT"]
        result[self.conf.usage_field] = self.get_udf_call_example(
            result, input_type=info["SCRIPT_INPUT_TYPE"], func_name=info["SCRIPT_NAME"]
        )
        return result
